from __future__ import division, print_function

from unittest import TestCase

import numpy as np
import pyroomacoustics as pra
from pyroomacoustics.transform import STFT
from scipy.signal import fftconvolve

"""
We create a signal, a simple filter and compute their convolution.

Then we test STFT block procesing with and without overlap,
and with and without filtering. Simulating a case of real-time
block processing.
"""

# test parameters
tol = 5e-6
np.random.seed(0)
D = 4
transform = "numpy"  # 'numpy', 'pyfftw', or 'mkl'

# filter to apply
h_len = 99
h = np.ones((h_len, D))
h /= np.linalg.norm(h, axis=0)

# test signal (noise)
x = np.random.randn(100000, D).astype(np.float32)

# convolved signal
y = np.zeros((x.shape[0] + h_len - 1, x.shape[1]))
for i in range(x.shape[1]):
    y[:, i] = fftconvolve(x[:, i], h[:, i])


def incorrect_input_size(D, num_frames):

    if D == 1:
        x_local = x[:, 0]
    else:
        x_local = x[:, :D]

    # parameters
    block_size = 512
    hop = block_size

    # create STFT object
    stft = STFT(
        block_size, hop=hop, channels=D, transform=transform, num_frames=num_frames
    )

    try:  # passing more frames than 'hop'
        stft.analysis(x_local)
        computed = False
    except:
        computed = True

    return computed


def no_overlap_no_filter(D, num_frames=1, fixed_memory=False, streaming=True):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at
                    a time
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:, 0]
    else:
        x_local = x[:, :D]

    # parameters
    block_size = 512  # make sure the FFT size is a power of 2
    hop = block_size  # no overlap
    if not streaming:
        num_samples = (num_frames - 1) * hop + block_size
        x_local = x_local[
            :num_samples,
        ]

    # Create the STFT object
    if fixed_memory:
        stft = STFT(
            block_size,
            hop=hop,
            channels=D,
            transform=transform,
            num_frames=num_frames,
            streaming=streaming,
        )
    else:
        stft = STFT(
            block_size, hop=hop, channels=D, transform=transform, streaming=streaming
        )

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if streaming:

        n = 0
        hop_frames = hop * num_frames
        # process the signals while full blocks are available
        while x_local.shape[0] - n > hop_frames:
            stft.analysis(
                x_local[
                    n : n + hop_frames,
                ]
            )
            processed_x[
                n : n + hop_frames,
            ] = stft.synthesis()
            n += hop_frames

    else:

        stft.analysis(x_local)
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

    error = np.max(
        np.abs(
            x_local[
                :n,
            ]
            - processed_x[
                :n,
            ]
        )
    )

    return error


def with_arbitrary_overlap_synthesis_window(
    D, num_frames=1, fixed_memory=False, streaming=True, overlap=0.5
):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at
                    a time
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:, 0]
    else:
        x_local = x[:, :D]

    # parameters
    block_size = 512  # make sure the FFT size is a power of 2
    hop = int((1 - overlap) * block_size)  # quarter overlap
    if not streaming:
        num_samples = (num_frames - 1) * hop + block_size
        x_local = x_local[
            :num_samples,
        ]

    analysis_window = pra.hann(block_size)
    synthesis_window = pra.transform.stft.compute_synthesis_window(analysis_window, hop)

    # Create the STFT object
    if fixed_memory:
        stft = STFT(
            block_size,
            hop=hop,
            channels=D,
            transform=transform,
            num_frames=num_frames,
            analysis_window=analysis_window,
            synthesis_window=synthesis_window,
            streaming=streaming,
        )
    else:
        stft = STFT(
            block_size,
            hop=hop,
            channels=D,
            analysis_window=analysis_window,
            synthesis_window=synthesis_window,
            transform=transform,
            streaming=streaming,
        )

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if streaming:

        n = 0
        hop_frames = hop * num_frames
        # process the signals while full blocks are available
        while x_local.shape[0] - n > hop_frames:
            stft.analysis(
                x_local[
                    n : n + hop_frames,
                ]
            )
            processed_x[
                n : n + hop_frames,
            ] = stft.synthesis()
            n += hop_frames

        error = np.max(
            np.abs(
                x_local[
                    : n - block_size + hop,
                ]
                - processed_x[
                    block_size - hop : n,
                ]
            )
        )

        if 20 * np.log10(error) > -10:
            import matplotlib.pyplot as plt

            if x_local.ndim == 1:
                plt.plot(x_local[: n - block_size + hop])
                plt.plot(processed_x[block_size - hop : n])
            else:
                plt.plot(x_local[: n - block_size + hop, 0])
                plt.plot(processed_x[block_size - hop : n, 0])
            plt.show()

    else:

        stft.analysis(x_local)
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

        L = block_size - hop
        error = np.max(
            np.abs(
                x_local[
                    L:-L,
                ]
                - processed_x[
                    L:,
                ]
            )
        )

        if 20 * np.log10(error) > -10:
            import matplotlib.pyplot as plt

            if x_local.ndim == 1:
                plt.plot(x_local[L:-L])
                plt.plot(processed_x[L:])
            else:
                plt.plot(x_local[L:-L, 0])
                plt.plot(processed_x[L:, 0])
            plt.show()

    return error


def no_overlap_with_filter(D, num_frames=1, fixed_memory=False, streaming=True):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at
                    a time
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:, 0]
        y_local = y[:, 0]
        h_local = h[:, 0]
    else:
        x_local = x[:, :D]
        y_local = y[:, :D]
        h_local = h[:, :D]

    # parameters
    block_size = 512 - h_len + 1  # make sure the FFT size is a power of 2
    hop = block_size  # no overlap
    if not streaming:
        num_samples = (num_frames - 1) * hop + block_size
        x_local = x_local[
            :num_samples,
        ]

    # Create the STFT object
    if fixed_memory:
        stft = STFT(
            block_size,
            hop=hop,
            channels=D,
            transform=transform,
            num_frames=num_frames,
            streaming=streaming,
        )
    else:
        stft = STFT(
            block_size, hop=hop, channels=D, transform=transform, streaming=streaming
        )

    # setup the filter
    stft.set_filter(h_local, zb=h_len - 1)

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if not streaming:

        stft.analysis(x_local)
        stft.process()
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

    else:

        n = 0
        hop_frames = hop * num_frames
        # process the signals while full blocks are available
        while x_local.shape[0] - n > hop_frames:
            stft.analysis(
                x_local[
                    n : n + hop_frames,
                ]
            )
            stft.process()  # apply the filter
            processed_x[
                n : n + hop_frames,
            ] = stft.synthesis()
            n += hop_frames

    error = np.max(
        np.abs(
            y_local[
                :n,
            ]
            - processed_x[
                :n,
            ]
        )
    )

    return error


def with_half_overlap_no_filter(D, num_frames=1, fixed_memory=False, streaming=True):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at
                    a time
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:, 0]
    else:
        x_local = x[:, :D]

    # parameters
    block_size = 512  # make sure the FFT size is a power of 2
    hop = block_size // 2  # half overlap
    window = pra.hann(block_size)  # the analysis window
    if not streaming:
        num_samples = (num_frames - 1) * hop + block_size
        x_local = x_local[
            :num_samples,
        ]

    # Create the STFT object
    if fixed_memory:
        stft = STFT(
            block_size,
            hop=hop,
            channels=D,
            transform=transform,
            num_frames=num_frames,
            analysis_window=window,
            streaming=streaming,
        )
    else:
        stft = STFT(
            block_size,
            hop=hop,
            channels=D,
            transform=transform,
            analysis_window=window,
            streaming=streaming,
        )

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if not streaming:

        stft.analysis(x_local)
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

        error = np.max(
            np.abs(
                x_local[
                    block_size - hop : n - hop,
                ]
                - processed_x[
                    block_size - hop : n - hop,
                ]
            )
        )

    else:

        n = 0
        hop_frames = hop * num_frames
        # process the signals while full blocks are available
        while x_local.shape[0] - n > hop_frames:
            stft.analysis(
                x_local[
                    n : n + hop_frames,
                ]
            )
            processed_x[
                n : n + hop_frames,
            ] = stft.synthesis()
            n += hop_frames

        error = np.max(
            np.abs(
                x_local[
                    : n - hop,
                ]
                - processed_x[
                    hop:n,
                ]
            )
        )

    return error


def with_half_overlap_with_filter(D, num_frames=1, fixed_memory=False, streaming=True):
    """
    D             - number of channels
    num_frames    - how many frames to process, None will process one frame at
                    a time
    fixed_memory  - whether to enforce checks for size (real-time consideration)
    streaming     - whether or not to stitch between frames
    """

    if D == 1:
        x_local = x[:, 0]
        y_local = y[:, 0]
        h_local = h[:, 0]
    else:
        x_local = x[:, :D]
        y_local = y[:, :D]
        h_local = h[:, :D]

    # parameters
    block_size = 512 - h_len + 1  # make sure the FFT size is a power of 2
    hop = block_size // 2  # half overlap
    window = pra.hann(block_size)  # the analysis window
    if not streaming:
        num_samples = (num_frames - 1) * hop + block_size
        x_local = x_local[
            :num_samples,
        ]

    # Create the STFT object
    if fixed_memory:
        stft = STFT(
            block_size,
            hop=hop,
            channels=D,
            transform=transform,
            num_frames=num_frames,
            analysis_window=window,
            streaming=streaming,
        )
    else:
        stft = STFT(
            block_size,
            hop=hop,
            channels=D,
            transform=transform,
            analysis_window=window,
            streaming=streaming,
        )

    # setup the filter
    stft.set_filter(h_local, zb=h_len - 1)

    # collect the processed blocks
    processed_x = np.zeros(x_local.shape)

    if not streaming:

        stft.analysis(x_local)
        stft.process()
        processed_x = stft.synthesis()
        n = processed_x.shape[0]

        error = np.max(
            np.abs(
                y_local[
                    block_size : n - block_size,
                ]
                - processed_x[
                    block_size : n - block_size,
                ]
            )
        )

    else:

        n = 0
        hop_frames = hop * num_frames
        # process the signals while full blocks are available
        while x_local.shape[0] - n > hop_frames:
            stft.analysis(
                x_local[
                    n : n + hop_frames,
                ]
            )
            stft.process()  # apply the filter
            processed_x[
                n : n + hop_frames,
            ] = stft.synthesis()
            n += hop_frames

        error = np.max(
            np.abs(
                y_local[
                    : n - hop,
                ]
                - processed_x[
                    hop:n,
                ]
            )
        )

        # if D==1:
        #     import matplotlib.pyplot as plt
        #     plt.figure()
        #     plt.plot(y_local)
        #     plt.plot(processed_x)
        #     plt.show()

    return error


def call_all_stft_tests(num_frames=1, fixed_memory=False, streaming=True, overlap=True):

    error = no_overlap_no_filter(1, num_frames, fixed_memory, streaming)
    print("no overlap, no filter, mono             : %0.0f dB" % (20 * np.log10(error)))

    error = no_overlap_no_filter(D, num_frames, fixed_memory, streaming)
    print("no overlap, no filter, multichannel     : %0.0f dB" % (20 * np.log10(error)))

    error = no_overlap_with_filter(1, num_frames, fixed_memory, streaming)
    print("no overlap, with filter, mono           : %0.0f dB" % (20 * np.log10(error)))

    error = no_overlap_with_filter(D, num_frames, fixed_memory, streaming)
    print("no overlap, with filter, multichannel   : %0.0f dB" % (20 * np.log10(error)))

    if overlap:
        error = with_half_overlap_no_filter(1, num_frames, fixed_memory, streaming)
        print(
            "half overlap, no filter, mono           : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            1, num_frames, fixed_memory, streaming, overlap=0.5
        )
        print(
            "half overlap, no filter, with synthesis windows, mono : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            D, num_frames, fixed_memory, streaming, overlap=0.5
        )
        print(
            "half overlap, no filter, with synthesis windows, multichannel : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            1, num_frames, fixed_memory, streaming, overlap=0.75
        )
        print(
            "3/4 overlap, no filter, with synthesis windows, mono : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            D, num_frames, fixed_memory, streaming, overlap=0.75
        )
        print(
            "3/4 overlap, no filter, with synthesis windows, multichannel : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            1, num_frames, fixed_memory, streaming, overlap=0.84
        )
        print(
            "84/100 overlap, no filter, with synthesis windows, mono : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            D, num_frames, fixed_memory, streaming, overlap=0.84
        )
        print(
            "84/100 overlap, no filter, with synthesis windows, multichannel : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            1, num_frames, fixed_memory, streaming, overlap=0.26
        )
        print(
            "26/100 overlap, no filter, with synthesis windows, mono : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            D, num_frames, fixed_memory, streaming, overlap=0.26
        )
        print(
            "26/100 overlap, no filter, with synthesis windows, multichannel : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            1, num_frames, fixed_memory, streaming, overlap=7 / 8
        )
        print(
            "7/8 overlap, no filter, with synthesis windows, mono : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            D, num_frames, fixed_memory, streaming, overlap=7 / 8
        )
        print(
            "7/8 overlap, no filter, with synthesis windows, multichannel : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            1, num_frames, fixed_memory, streaming, overlap=0.25
        )
        print(
            "1/4 overlap, no filter, with synthesis windows, mono : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_arbitrary_overlap_synthesis_window(
            D, num_frames, fixed_memory, streaming, overlap=0.25
        )
        print(
            "1/4 overlap, no filter, with synthesis windows, multichannel : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_half_overlap_no_filter(D, num_frames, fixed_memory, streaming)
        print(
            "half overlap, no filter, multichannel   : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_half_overlap_with_filter(1, num_frames, fixed_memory, streaming)
        print(
            "half overlap, with filter, mono         : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_half_overlap_with_filter(D, num_frames, fixed_memory, streaming)
        print(
            "half overlap, with filter, multichannel : %0.0f dB"
            % (20 * np.log10(error))
        )

        error = with_half_overlap_with_filter(D, num_frames, fixed_memory, streaming)
        print(
            "half overlap, with filter, multichannel : %0.0f dB"
            % (20 * np.log10(error))
        )
    print()


class TestSTFT(TestCase):
    def test_incorrect_input_check(self):
        result = incorrect_input_size(1, 100)
        self.assertTrue(result)
        result = incorrect_input_size(D, 100)
        self.assertTrue(result)

    def test_no_overlap_no_filter_mono(self):
        error = no_overlap_no_filter(
            D=1, num_frames=1, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=1, num_frames=50, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=1, num_frames=1, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=1, num_frames=50, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=1, num_frames=1, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=1, num_frames=50, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=1, num_frames=1, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=1, num_frames=50, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)

    def test_no_overlap_no_filter_multichannel(self):
        error = no_overlap_no_filter(
            D=D, num_frames=1, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=D, num_frames=50, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=D, num_frames=1, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=D, num_frames=50, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=D, num_frames=1, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=D, num_frames=50, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=D, num_frames=1, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_no_filter(
            D=D, num_frames=50, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)

    def test_no_overlap_with_filter_mono(self):
        error = no_overlap_with_filter(
            D=1, num_frames=1, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=1, num_frames=50, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=1, num_frames=1, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=1, num_frames=50, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=1, num_frames=1, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=1, num_frames=50, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=1, num_frames=1, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=1, num_frames=50, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)

    def test_no_overlap_with_filter_multichannel(self):
        error = no_overlap_with_filter(
            D=D, num_frames=1, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=D, num_frames=50, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=D, num_frames=1, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=D, num_frames=50, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=D, num_frames=1, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=D, num_frames=50, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=D, num_frames=1, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)
        error = no_overlap_with_filter(
            D=D, num_frames=50, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)

    def test_with_half_overlap_no_filter_mono(self):
        error = with_half_overlap_no_filter(
            D=1, num_frames=1, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=1, num_frames=50, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=1, num_frames=1, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=1, num_frames=50, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=1, num_frames=50, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=1, num_frames=50, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)

    def test_with_half_overlap_no_filter_multichannel(self):
        error = with_half_overlap_no_filter(
            D=D, num_frames=1, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=D, num_frames=50, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=D, num_frames=1, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=D, num_frames=50, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=D, num_frames=50, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_no_filter(
            D=D, num_frames=50, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)

    def test_with_half_overlap_with_filter_mono(self):
        error = with_half_overlap_with_filter(
            D=1, num_frames=1, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=1, num_frames=50, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=1, num_frames=1, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=1, num_frames=50, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=1, num_frames=50, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=1, num_frames=50, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)

    def test_with_half_overlap_with_filter_multichannel(self):
        error = with_half_overlap_with_filter(
            D=D, num_frames=1, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=D, num_frames=50, fixed_memory=False, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=D, num_frames=1, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=D, num_frames=50, fixed_memory=True, streaming=True
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=D, num_frames=50, fixed_memory=False, streaming=False
        )
        self.assertTrue(error < tol)
        error = with_half_overlap_with_filter(
            D=D, num_frames=50, fixed_memory=True, streaming=False
        )
        self.assertTrue(error < tol)

    def test_with_arbitrary_overlap_synthesis_window_multichannel(self):
        overlaps = [0.5, 0.75, 0.84, 0.26, 7 / 8, 0.25]
        for overlap in overlaps:
            error = with_arbitrary_overlap_synthesis_window(
                D, num_frames=1, fixed_memory=False, streaming=True, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                D, num_frames=50, fixed_memory=False, streaming=True, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                D, num_frames=1, fixed_memory=True, streaming=True, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                D, num_frames=50, fixed_memory=True, streaming=True, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                D, num_frames=50, fixed_memory=False, streaming=False, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                D, num_frames=50, fixed_memory=True, streaming=False, overlap=overlap
            )
            self.assertTrue(error < tol)

    def test_with_arbitrary_overlap_synthesis_window_mono(self):
        overlaps = [0.5, 0.75, 0.84, 0.26, 7 / 8, 0.25]
        for overlap in overlaps:
            error = with_arbitrary_overlap_synthesis_window(
                1, num_frames=1, fixed_memory=False, streaming=True, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                1, num_frames=50, fixed_memory=False, streaming=True, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                1, num_frames=1, fixed_memory=True, streaming=True, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                1, num_frames=50, fixed_memory=True, streaming=True, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                1, num_frames=50, fixed_memory=False, streaming=False, overlap=overlap
            )
            self.assertTrue(error < tol)
            error = with_arbitrary_overlap_synthesis_window(
                1, num_frames=50, fixed_memory=True, streaming=False, overlap=overlap
            )
            self.assertTrue(error < tol)


if __name__ == "__main__":

    print()
    print("TEST INFO")
    print("-------------------------------------------------------------")
    print("Max error in dB for randomly generated signal of %d samples." % len(x))
    print("Multichannel corresponds to %d channels." % D)
    print("-------------------------------------------------------------")
    print()

    print("---ONE FRAME, STREAMING, NOT FIXED MEMORY")
    call_all_stft_tests(num_frames=1, fixed_memory=False, streaming=True)

    print("---MULTIPLE FRAMES, STREAMING, NOT FIXED MEMORY")
    call_all_stft_tests(num_frames=50, fixed_memory=False, streaming=True)

    print("---ONE FRAME, STREAMING, FIXED MEMORY")
    num_frames = 1
    result = incorrect_input_size(1, num_frames)
    print("incorrect input size, mono              :", result)
    result = incorrect_input_size(D, num_frames)
    print("incorrect input size, multichannel      :", result)
    call_all_stft_tests(num_frames=num_frames, fixed_memory=True, streaming=True)

    print("---MULTIPLE FRAME, STREAMING, FIXED MEMORY")
    num_frames = 50
    result = incorrect_input_size(1, num_frames)
    print("incorrect input size, mono              :", result)
    result = incorrect_input_size(D, num_frames)
    print("incorrect input size, multichannel      :", result)
    call_all_stft_tests(num_frames=num_frames, fixed_memory=True, streaming=True)

    print("---ONE FRAME, NON-STREAMING, NOT FIXED MEMORY")
    call_all_stft_tests(
        num_frames=1, fixed_memory=False, streaming=False, overlap=False
    )

    print("---MULTIPLE FRAMES, NON-STREAMING, NOT FIXED MEMORY")
    call_all_stft_tests(num_frames=50, fixed_memory=False, streaming=False)

    print("---ONE FRAME, NON-STREAMING, FIXED MEMORY")
    call_all_stft_tests(num_frames=1, fixed_memory=True, streaming=False, overlap=False)

    print("---MULTIPLE FRAMES, NON-STREAMING, FIXED MEMORY")
    call_all_stft_tests(num_frames=50, fixed_memory=True, streaming=False)



import numpy as np
import pyroomacoustics as pra


def test_soundsource_basic():

    the_signal = np.ones(10)
    source = pra.SoundSource([1.0, 1.0], signal=the_signal)

    the_signal_2 = np.zeros(10)
    source.add_signal(the_signal_2)


if __name__ == "__main__":

    test_soundsource_basic()



"""
This test attempts to check that the C module
runs fine. It was created to test Issue #22 on github
whereas the program would crash with
"Fatal Python Error: deallocating None"
after around 20000 calls to Room.compute_rir().

The bug was fixed by adding reference count increments
before returning Py_None in the C extension.

The test creates a random but fixed room and calls
Room.image_source_model() 25000 times.

If the C module is not installed (pure python
fallback version), then nothing is done.
"""
import numpy as np
import pyroomacoustics


def test_issue_22():

    np.random.seed(0)

    n_mics = 1
    n_src = 1
    n_times = 25000
    dim = 3
    mic_pos = np.random.rand(dim, n_mics)
    abs_coeff = 0.1
    e_abs = 1.0 - (1.0 - abs_coeff) ** 2
    fs = 16000
    wall_max_len = 15

    room_dim = np.random.rand(dim) * wall_max_len

    shoebox = pyroomacoustics.ShoeBox(
        room_dim,
        materials=pyroomacoustics.Material(e_abs),
        fs=fs,
        max_order=0,
    )

    src_pos = np.random.rand(dim, n_src) * room_dim[:, None]
    for src in src_pos.T:
        shoebox.add_source(src)

    shoebox.add_microphone_array(pyroomacoustics.MicrophoneArray(mic_pos, fs))

    for i in range(n_times):

        shoebox.image_source_model()

        if i != 0 and i % 1000 == 0:
            print(i)


if __name__ == "__main__":

    test_issue_22()



"""
Tests the mixing function of ``Room.simulate``
"""
import unittest

import numpy as np
import pyroomacoustics as pra

e_abs = 1.0 - (1.0 - 0.25) ** 2
room = pra.ShoeBox([9, 5, 4], fs=16000, materials=pra.Material(e_abs), max_order=15)

# three microphones
room.add_microphone_array(
    pra.MicrophoneArray(
        np.c_[
            [4.3, 2.1, 1.8],
            [2.5, 1.7, 1.2],
            [6.5, 3.9, 0.9],
        ],
        room.fs,
    )
)

# add two sources (1 target, 1 interferer)
room.add_source([3.1, 1.5, 2.1], signal=np.random.randn(room.fs * 5))
room.add_source([5.8, 3.3, 3.1], signal=np.random.randn(room.fs * 3), delay=1.0)

# the extra arguments are given in a dictionary
mix_kwargs = {
    "snr": 30,  # SNR target is 30 decibels
    "sir": 10,  # SIR target is 10 decibels
    "n_src": 2,
    "n_tgt": 1,
    "ref_mic": 1,
}


def callback_mix(premix, snr=0, sir=0, ref_mic=0, n_src=None, n_tgt=None):

    # first normalize all separate recording to have unit power at microphone one
    p_mic_ref = np.std(premix[:, ref_mic, :], axis=1)
    premix /= p_mic_ref[:, None, None]

    # now compute the power of interference signal needed to achieve desired SIR
    sigma_i = np.sqrt(10 ** (-sir / 10) / (n_src - n_tgt))
    premix[n_tgt:n_src, :, :] *= sigma_i

    # compute noise variance
    sigma_n = np.sqrt(10 ** (-snr / 10))

    # Mix down the recorded signals
    mix = np.sum(premix[:n_src, :], axis=0) + sigma_n * np.random.randn(
        *premix.shape[1:]
    )

    return mix


class TestRoomMix(unittest.TestCase):
    def test_mix(self):

        # Run the simulation
        premix = room.simulate(
            callback_mix=callback_mix,
            callback_mix_kwargs=mix_kwargs,
            return_premix=True,
        )
        mics_signals = room.mic_array.signals

        # recover the noise signal
        i_ref = mix_kwargs["ref_mic"]
        noise = mics_signals[i_ref, :] - np.sum(premix[:, i_ref, :], axis=0)
        noise_pwr = np.var(noise)

        premix_ref_pwr = np.var(premix[:, i_ref, :], axis=1)

        tgt_pwr = np.sum(premix_ref_pwr[: mix_kwargs["n_tgt"]])
        int_pwr = np.sum(premix_ref_pwr[mix_kwargs["n_tgt"] :])
        snr = 10 * np.log10(1.0 / noise_pwr)
        sir = 10 * np.log10(tgt_pwr / int_pwr)

        print("SNR", snr)
        print("SIR", sir)

        self.assertTrue(
            all(
                [
                    abs(snr - mix_kwargs["snr"]) < 5e-2,
                    abs(sir - mix_kwargs["sir"]) < 5e-2,
                ]
            )
        )


if __name__ == "__main__":
    unittest.main()



import warnings

import numpy
import numpy as np
import pyroomacoustics as pra

warnings.filterwarnings(action="ignore", category=UserWarning)


def random_room_ism(max_order=10, eps=1e-6, verbose=False):
    """
    Create a random shoebox room and compute the difference
    """

    # locations of stuff
    room_dim = np.random.randint(1, 101, size=3)
    src_loc = np.random.rand(3) * room_dim
    mic_loc = np.random.rand(3) * room_dim
    # too close is not good
    while np.linalg.norm(mic_loc - src_loc) < 0.05:
        mic_loc = np.random.rand(3) * room_dim

    # random list of materials
    materials = dict(
        zip(
            ["north", "south", "west", "east", "ceiling", "floor"],
            [pra.Material(x) for x in np.random.rand(6)],
        )
    )

    # shoebox room: working correctly
    room = pra.ShoeBox(room_dim, max_order=max_order, materials=materials)
    # general room: not working
    room2 = pra.Room(room.walls, max_order=max_order)

    room.add_source(src_loc)
    room2.add_source(src_loc)

    room.add_microphone(mic_loc)
    room2.add_microphone(mic_loc)

    room.image_source_model()
    room2.image_source_model()

    trans_shoebox = np.sort(room.sources[0].damping)
    trans_general = np.sort(room2.sources[0].damping)

    error = np.linalg.norm(trans_general - trans_shoebox)

    if verbose:
        print("error", np.linalg.norm(trans_shoebox - trans_general))

    assert error < eps


def test_ism_shoebox_vs_general(verbose=False):

    np.random.seed(0)
    n_repeat = 100
    max_order = 10
    eps = 5e-6

    for i in range(n_repeat):
        random_room_ism(max_order=max_order, eps=eps, verbose=verbose)


if __name__ == "__main__":
    test_ism_shoebox_vs_general(verbose=True)



import os
import platform

import numpy as np
from scipy.stats import binom as _binom
from scipy.stats import norm as _norm
from scipy.signal import hann

from .transform import stft


def median(x, alpha=None, axis=-1, keepdims=False):
    """
    Computes 95% confidence interval for the median.

    Parameters
    ----------
    x: array_like
        the data array
    alpha: float, optional
        the confidence level of the interval, confidence intervals are only computed
        when this argument is provided
    axis: int, optional
        the axis of the data on which to operate, by default the last axis

    Returns
    -------
    :tuple ``(float, [float, float])``
        This function returns ``(m, [le, ue])`` and the confidence interval is ``[m-le, m+ue]``.
    """

    # place the axis on which to compute median in first position
    xsw = np.moveaxis(x, axis, 0)

    # sort the array
    xsw = np.sort(xsw, axis=0)
    n = xsw.shape[0]

    if n % 2 == 1:
        # if n is odd, take central element
        m = xsw[
            n // 2,
        ]
    else:
        # if n is even, average the two central elements
        m = 0.5 * (
            xsw[
                n // 2 - 1,
            ]
            + xsw[
                n // 2,
            ]
        )

    if alpha is None:
        if keepdims:
            m = np.moveaxis(
                m[
                    np.newaxis,
                ],
                0,
                axis,
            )
        return m

    else:
        # bound for using the large n approximation
        clt_bound = max(10 / alpha, 10 / (2 - alpha))

        if n < clt_bound:
            # Get the bounds of the CI from the binomial distribution
            b = _binom(n, 0.5)
            j, k = int(b.ppf(alpha / 2) - 1), int(b.ppf(1 - alpha / 2) - 1)

            if b.cdf(k) - b.cdf(j) < 1 - alpha:
                k += 1

            # sanity check
            assert b.cdf(k) - b.cdf(j) >= 1 - alpha

            if j < 0:
                raise ValueError(
                    "Warning: Sample size is too small. No confidence interval found."
                )
            else:
                ci = np.array(
                    [
                        xsw[
                            j,
                        ]
                        - m,
                        xsw[
                            k,
                        ]
                        - m,
                    ]
                )

        else:
            # we use the Normal approximation for large sets
            norm = _norm()
            eta = norm.ppf(1 - alpha / 2)
            j = int(np.floor(0.5 * n - 0.5 * eta * np.sqrt(n))) - 1
            k = int(np.ceil(0.5 * n + 0.5 * eta * np.sqrt(n)))
            ci = np.array(
                [
                    xsw[
                        j,
                    ]
                    - m,
                    xsw[
                        k,
                    ]
                    - m,
                ]
            )

        if keepdims:
            m = np.moveaxis(
                m[
                    np.newaxis,
                ],
                0,
                axis,
            )
            if axis < 0:
                ci = np.moveaxis(
                    ci[
                        :,
                        np.newaxis,
                    ],
                    1,
                    axis,
                )
            else:
                ci = np.moveaxis(
                    ci[
                        :,
                        np.newaxis,
                    ],
                    1,
                    axis + 1,
                )

        return m, ci


# Simple mean squared error function
def mse(x1, x2):
    r"""
    A short hand to compute the mean-squared error of two signals.

    .. math::

       MSE = \\frac{1}{n}\sum_{i=0}^{n-1} (x_i - y_i)^2


    :arg x1: (ndarray)
    :arg x2: (ndarray)
    :returns: (float) The mean of the squared differences of x1 and x2.
    """
    return (np.abs(x1 - x2) ** 2).sum() / len(x1)


# Itakura-Saito distance function
def itakura_saito(x1, x2, sigma2_n, stft_L=128, stft_hop=128):

    P1 = np.abs(stft.analysis(x1, stft_L, stft_hop)) ** 2
    P2 = np.abs(stft(x2, stft_L, stft_hop)) ** 2

    VAD1 = P1.mean(axis=1) > 2 * stft_L**2 * sigma2_n
    VAD2 = P2.mean(axis=1) > 2 * stft_L**2 * sigma2_n
    VAD = np.logical_or(VAD1, VAD2)

    if P1.shape[0] != P2.shape[0] or P1.shape[1] != P2.shape[1]:
        raise ValueError("Error: Itakura-Saito requires both array to have same length")

    R = P1[VAD, :] / P2[VAD, :]

    IS = (R - np.log(R) - 1.0).mean(axis=1)

    return np.median(IS)


def snr(ref, deg):

    return np.sum(ref**2) / np.sum((ref - deg) ** 2)


# Perceptual Evaluation of Speech Quality for multiple files using multiple threads
def pesq(ref_file, deg_files, Fs=8000, swap=False, wb=False, bin="./bin/pesq"):
    """
    pesq_vals = pesq(ref_file, deg_files, sample_rate=None, bin='./bin/pesq'):
    Computes the perceptual evaluation of speech quality (PESQ) metric of a degraded
    file with respect to a reference file.  Uses the utility obtained from ITU
    P.862 http://www.itu.int/rec/T-REC-P.862-200511-I!Amd2/en

    :arg ref_file:    The filename of the reference file.
    :arg deg_files:   A list of degraded sound files names.
    :arg sample_rate: Sample rates of the sound files [8kHz or 16kHz, default 8kHz].
    :arg swap:        Swap byte orders (whatever that does is not clear to me) [default: False].
    :arg wb:          Use wideband algorithm [default: False].
    :arg bin:         Location of pesq executable [default: ./bin/pesq].

    :returns: (ndarray size 2xN) ndarray containing Raw MOS and MOS LQO in rows 0 and 1,
        respectively, and has one column per degraded file name in deg_files.
    """

    if isinstance(deg_files, str):
        deg_files = [deg_files]

    if platform.system() == "Windows":
        bin = bin + ".exe"

    if not os.path.isfile(ref_file):
        raise ValueError("Some file did not exist")
    for f in deg_files:
        if not os.path.isfile(f):
            raise ValueError("Some file did not exist")

    if Fs not in (8000, 16000):
        raise ValueError("sample rate must be 8000 or 16000")

    args = [bin, "+%d" % int(Fs)]

    if swap is True:
        args.append("+swap")

    if wb is True:
        args.append("+wb")

    args.append(ref_file)

    # array to receive all output values
    pesq_vals = np.zeros((2, len(deg_files)))

    # launch pesq for each degraded file in a different process
    import subprocess

    pipes = [
        subprocess.Popen(args + [deg], stdout=subprocess.PIPE) for deg in deg_files
    ]
    states = np.ones(len(pipes), dtype=np.bool)

    # Recover output as the processes finish
    while states.any():

        for i, p in enumerate(pipes):
            if states[i] == True and p.poll() is not None:
                states[i] = False
                out = p.stdout.readlines()
                last_line = out[-1][:-2]

                if wb is True:
                    if not last_line.startswith("P.862.2 Prediction"):
                        raise ValueError(last_line)
                    pesq_vals[:, i] = np.array([0, float(last_line.split()[-1])])
                else:
                    if not last_line.startswith("P.862 Prediction"):
                        raise ValueError(last_line)
                    pesq_vals[:, i] = np.array(map(float, last_line.split()[-2:]))

    return pesq_vals


def sweeping_echo_measure(rir, fs, t_min=0, t_max=0.5, fb=400):
    """
    Measure of sweeping echo in RIR obtained from image-source method.
    A higher value indicates less sweeping echoes

    For details see : De Sena et al. "On the modeling of rectangular geometries in
    room acoustic simulations", IEEE TASLP, 2015


    Parameters
    ----------

    rir:    RIR signal from ISM (mono).
    fs:     sampling frequency.
    t_min:  TYPE, optional
            Minimum time window. The default is 0.
    t_max:  TYPE, optional
            Maximum time window. The default is 0.5.
    fb:     TYPE, optional
            Mask bandwidth. The default is 400.

    Returns
    -------
    sweeping spectrum flatness (ssf)

    """

    # some default values
    fmin = 50
    fmax = 0.9 * fs / 2

    # STFT parameters
    fft_size = 512  # fft size for analysis
    fft_hop = 256  # hop between analysis frame
    fft_zp = 2**12 - fft_size  # zero padding
    analysis_window = hann(fft_size)

    # calculate stft
    S = stft.analysis(rir, fft_size, fft_hop, win=analysis_window, zp_back=fft_zp)

    (nFrames, nFreqs) = np.shape(S)
    nFreqs = int(nFreqs / 2)

    # ignore negative frequencies
    S = S[:, :nFreqs]

    timeSlice = np.arange(0, nFrames) * fft_hop / fs
    assert nFrames == len(timeSlice)
    freqSlice = np.linspace(0, fs / 2, nFreqs)
    assert nFreqs == len(freqSlice)

    # get time-frequency grid
    (t_mesh, f_mesh) = np.meshgrid(timeSlice, freqSlice)

    bmin = int(np.floor(nFreqs / fs * fmin))
    bmax = int(np.ceil(nFreqs / fs * fmax))

    Phi = np.zeros(np.shape(S))

    # normalize spectrogram to make energy identical in each bin
    for k in range(nFrames):
        norm_factor = np.sum(np.power(np.abs(S[k, bmin:bmax]), 2))
        Phi[k, :] = np.abs(S[k, :]) / np.sqrt(norm_factor)

    # slope values
    nCoeffs = 500
    coeffs = np.linspace(5000, 150000, nCoeffs)
    ss = np.zeros(nCoeffs)

    # loop through different slope values
    for k in range(nCoeffs):

        # get masks
        a = coeffs[k]

        maskTime = np.logical_and(t_mesh > t_min, t_mesh < t_max)

        maskFreq = np.logical_and(
            f_mesh > t_mesh * a - fb / 2, f_mesh < t_mesh * a + fb / 2
        )

        boolMask = np.logical_and(maskTime, maskFreq)

        ss[k] = np.mean(Phi[boolMask.T])

    # calculate spectral flatness
    ssf = np.exp(np.mean(np.log(np.abs(ss)))) / np.mean(np.abs(ss))

    return ssf



# Test of Wall intersection routines
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import numpy as np
import pyroomacoustics as pra

eps = 1e-6

cases = {
    "3d_valid": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[-1, -1, -1], [1, 1, 1]]),
        "intersect": np.array([1 / 3, 1 / 3, 1 / 3]),
        "type": pra.libroom.WALL_ISECT_VALID,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_endpt_1": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[1 / 3, 1 / 3, 1 / 3], [1, 1, 1]]),
        "intersect": np.array([1 / 3, 1 / 3, 1 / 3]),
        "type": pra.libroom.WALL_ISECT_VALID_ENDPT,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_endpt_2": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[1, 1, 1], [1 / 3, 1 / 3, 1 / 3]]),
        "intersect": np.array([1 / 3, 1 / 3, 1 / 3]),
        "type": pra.libroom.WALL_ISECT_VALID_ENDPT,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_bndry": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[1, 0, 1], [-1, 0, 1]]),
        "intersect": np.array([0, 0, 1]),
        "type": pra.libroom.WALL_ISECT_VALID_BNDRY,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_vertex": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[-2, 1, 1], [0, 1, 1]]),
        "intersect": np.array([-1, 1, 1]),
        "type": pra.libroom.WALL_ISECT_VALID_BNDRY,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_endpt_bndry_1": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[0, 0, 1], [-1, 0, 1]]),
        "intersect": np.array([0, 0, 1]),
        "type": pra.libroom.WALL_ISECT_VALID_BNDRY | pra.libroom.WALL_ISECT_VALID_ENDPT,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_endpt_bndry_2": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[1, 0, 1], [0, 0, 1]]),
        "intersect": np.array([0, 0, 1]),
        "type": pra.libroom.WALL_ISECT_VALID_BNDRY | pra.libroom.WALL_ISECT_VALID_ENDPT,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_vertex_endpt_1": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[-2, 1, 1], [-1, 1, 1]]),
        "intersect": np.array([-1, 1, 1]),
        "type": pra.libroom.WALL_ISECT_VALID_BNDRY | pra.libroom.WALL_ISECT_VALID_ENDPT,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_vertex_endpt_2": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[-1, 1, 1], [0, 1, 1]]),
        "intersect": np.array([-1, 1, 1]),
        "type": pra.libroom.WALL_ISECT_VALID_BNDRY | pra.libroom.WALL_ISECT_VALID_ENDPT,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_none_1": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[-10, -5, -3], [-2, -1, -3]]),
        "intersect": None,
        "type": pra.libroom.WALL_ISECT_NONE,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "3d_none_2": {
        "corners": np.array(
            [
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ]
        ),
        "seg": np.array([[-1, -1, 2], [1, 1, 2]]),
        "intersect": None,
        "type": pra.libroom.WALL_ISECT_NONE,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_valid": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[-1, 1], [1, 1]]),
        "intersect": np.array([1 / 3, 1]),
        "type": pra.libroom.WALL_ISECT_VALID,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_endpt_1": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[1 / 3, 1], [1, 1]]),
        "intersect": np.array([1 / 3, 1]),
        "type": pra.libroom.WALL_ISECT_VALID_ENDPT,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_endpt_2": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[-1, 1], [1 / 3, 1]]),
        "intersect": np.array([1 / 3, 1]),
        "type": pra.libroom.WALL_ISECT_VALID_ENDPT,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_bndry_1": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[0, 2], [2, 2]]),
        "intersect": np.array([1, 2]),
        "type": pra.libroom.WALL_ISECT_VALID_BNDRY,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_bndry_2": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[-2, -1], [0, -1]]),
        "intersect": np.array([-1, -1]),
        "type": pra.libroom.WALL_ISECT_VALID_BNDRY,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_bndry_endpt_1": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[-1, -1], [0, 1]]),
        "intersect": np.array([-1, -1]),
        "type": pra.libroom.WALL_ISECT_VALID_ENDPT | pra.libroom.WALL_ISECT_VALID_BNDRY,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_bndry_endpt_2": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[-2, -1], [-1, -1]]),
        "intersect": np.array([-1, -1]),
        "type": pra.libroom.WALL_ISECT_VALID_ENDPT | pra.libroom.WALL_ISECT_VALID_BNDRY,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_none_1": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[-2, 3], [3, 3]]),
        "intersect": None,
        "type": pra.libroom.WALL_ISECT_NONE,
        "absorption": [0.2],
        "scattering": [0.1],
    },
    "2d_none_2": {
        "corners": np.array(
            [  # [-1,-1] -> [1/3,1] -> [1, 2]
                [-1, 1],
                [-1, 2],
            ]
        ),
        "seg": np.array([[-4, 1], [-1, 1]]),
        "intersect": None,
        "type": pra.libroom.WALL_ISECT_NONE,
        "absorption": [0.2],
        "scattering": [0.1],
    },
}


def run_intersect(lbl):

    case = cases[lbl]
    wall = pra.wall_factory(case["corners"], case["absorption"], case["scattering"])
    p1, p2 = case["seg"]

    isect_exp = case["intersect"]
    r_exp = case["type"]
    isect = np.zeros(wall.dim, dtype=np.float32)
    ret = wall.intersection(p1, p2, isect)

    assert ret == r_exp, "Wall intersect: expects={} got={}".format(r_exp, ret)

    if r_exp != pra.libroom.WALL_ISECT_NONE:
        err = np.linalg.norm(isect - case["intersect"])
        assert err < eps, "Wall intersect: err={} (expects={} got={})".format(
            err, isect_exp, isect
        )


def test_3d_valid():
    run_intersect("3d_valid")


def test_3d_endpt_1():
    run_intersect("3d_endpt_1")


def test_3d_endpt_2():
    run_intersect("3d_endpt_2")


def test_3d_bndry():
    run_intersect("3d_bndry")


def test_3d_vertex():
    run_intersect("3d_vertex")


def test_3d_endpt_bndry_1():
    run_intersect("3d_endpt_bndry_1")


def test_3d_endpt_bndry_2():
    run_intersect("3d_endpt_bndry_2")


def test_3d_vertex_endpt_1():
    run_intersect("3d_vertex_endpt_1")


def test_3d_vertex_endpt_2():
    run_intersect("3d_vertex_endpt_2")


def test_3d_none_1():
    run_intersect("3d_none_1")


def test_3d_none_2():
    run_intersect("3d_none_2")


def test_2d_valid():
    run_intersect("2d_valid")


def test_2d_endpt_1():
    run_intersect("2d_endpt_1")


def test_2d_endpt_2():
    run_intersect("2d_endpt_2")


def test_2d_bndry_1():
    run_intersect("2d_bndry_1")


def test_2d_bndry_2():
    run_intersect("2d_bndry_2")


def test_2d_bndry_endpt_1():
    run_intersect("2d_bndry_endpt_1")


def test_2d_bndry_endpt_2():
    run_intersect("2d_bndry_endpt_2")


def test_2d_none_1():
    run_intersect("2d_none_1")


def test_2d_none_2():
    run_intersect("2d_none_2")


if __name__ == "__main__":

    test_3d_valid()
    test_3d_endpt_1()
    test_3d_endpt_2()
    test_3d_bndry()
    test_3d_vertex()
    test_3d_endpt_bndry_1()
    test_3d_endpt_bndry_2()
    test_3d_vertex_endpt_1()
    test_3d_vertex_endpt_2()
    test_3d_none_1()
    test_3d_none_2()

    test_2d_valid()
    test_2d_endpt_1()
    test_2d_endpt_2()
    test_2d_bndry_1()
    test_2d_bndry_2()
    test_2d_bndry_endpt_1()
    test_2d_bndry_endpt_2()
    test_2d_none_1()
    test_2d_none_2()



"""
Routines to perform grid search on the sphere
"""
from __future__ import division, print_function, absolute_import

import numpy as np
import scipy.spatial as sp  # import ConvexHull, SphericalVoronoi

from abc import ABCMeta, abstractmethod

from .detect_peaks import detect_peaks
from .utils import great_circ_dist


class Grid:
    """
    This is an abstract class with attributes and methods for grids

    Parameters
    ----------
    n_points: int
        the number of points on the grid
    """

    __metaclass__ = ABCMeta

    def __init__(self, n_points):

        self.n_points = n_points
        self.values = None
        self.cartesian = np.zeros((3, n_points))
        self.spherical = np.zeros((2, n_points))
        self.x = self.cartesian[0, :]
        self.y = self.cartesian[1, :]
        self.z = self.cartesian[2, :]
        self.azimuth = self.spherical[0, :]
        self.colatitude = self.spherical[1, :]
        self.dim = 0
        self.values = None

    @abstractmethod
    def apply(self, func, spherical=False):
        return NotImplemented

    @abstractmethod
    def find_peaks(self, k=1):
        return NotImplemented

    def set_values(self, vals):

        vals = np.array(vals)

        if vals.ndim == 0:
            self.values = np.ones(self.n_points) * vals

        else:
            if vals.shape != (self.n_points,):
                raise ValueError(
                    "Values should be a scalar or a 1D ndarray of the grid size."
                )

            self.values = vals


class GridCircle(Grid):
    """
    Creates a grid on the circle.

    Parameters
    ----------
    n_points: int, optional
        The number of uniformly spaced points in the grid.
    azimuth: ndarray, optional
        An array of azimuth (in radians) to use for grid locations. Overrides n_points.
    """

    def __init__(self, n_points=360, azimuth=None):
        if azimuth is not None:

            if azimuth.ndim != 1:
                raise ValueError("Azimuth must be a 1D ndarray.")

            azimuth = np.sort(azimuth)

            n_points = azimuth.shape[0]

        else:
            azimuth = np.linspace(0, 2 * np.pi, n_points, endpoint=False)

        Grid.__init__(self, n_points)

        self.dim = 2

        # spherical coordinates
        self.azimuth[:] = azimuth
        self.colatitude[:] = np.pi / 2  # Fix colatitude to ecuador

        # cartesian coordinates
        self.x[:] = np.cos(azimuth)
        self.y[:] = np.sin(azimuth)

    def apply(self, func, spherical=False):

        if spherical:
            self.values = func(self.azimuth)
        else:
            self.values = func(self.x, self.y)

    def find_peaks(self, k=1):

        # make circular
        val_ext = np.append(self.values, self.values[:10])

        # run peak finding
        indexes = detect_peaks(val_ext, show=False) % self.n_points
        candidates = np.unique(indexes)  # get rid of duplicates, if any

        # Select k largest
        peaks = self.values[candidates]
        max_idx = np.argsort(peaks)[-k:]

        # return the indices of peaks found
        return candidates[max_idx]

    def plot(self, mark_peaks=0):

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="polar")

        pts = np.append(self.azimuth, self.azimuth[0])
        vals = np.append(self.values, self.values[0])

        ax.plot(pts, vals, "-")

        if mark_peaks > 0:

            idx = self.find_peaks(k=mark_peaks)

            ax.plot(pts[idx], vals[idx], "ro")


class GridSphere(Grid):
    """
    This function computes nearly equidistant points on the sphere
    using the fibonacci method

    Parameters
    ----------
    n_points: int
        The number of points to sample
    spherical_points: ndarray, optional
        A 2 x n_points array of spherical coordinates with azimuth in
        the top row and colatitude in the second row. Overrides n_points.

    References
    ----------
    http://lgdv.cs.fau.de/uploads/publications/spherical_fibonacci_mapping.pdf
    http://stackoverflow.com/questions/9600801/evenly-distributing-n-points-on-a-sphere
    """

    def __init__(self, n_points=1000, spherical_points=None):
        if spherical_points is not None:

            if spherical_points.ndim != 2 or spherical_points.shape[0] != 2:
                raise ValueError("spherical_points must be a 2D array with two rows.")

            n_points = spherical_points.shape[1]

        # Parent constructor
        Grid.__init__(self, n_points)

        self.dim = 3

        if spherical_points is not None:
            # If a list of points was provided, use it

            self.spherical[:, :] = spherical_points

            # transform to cartesian coordinates
            self.x[:] = np.cos(self.azimuth) * np.sin(self.colatitude)
            self.y[:] = np.sin(self.azimuth) * np.sin(self.colatitude)
            self.z[:] = np.cos(self.colatitude)

        else:
            # If no list was provided, samples points on the sphere
            # as uniformly as possible

            # Fibonnaci sampling
            offset = 2.0 / n_points
            increment = np.pi * (3.0 - np.sqrt(5.0))

            self.z[:] = (np.arange(n_points) * offset - 1) + offset / 2
            rho = np.sqrt(1.0 - self.z**2)

            phi = np.arange(n_points) * increment

            self.x[:] = np.cos(phi) * rho
            self.y[:] = np.sin(phi) * rho

            # Create convenient arrays
            # to access both in cartesian and spherical coordinates
            self.azimuth[:] = np.arctan2(self.y, self.x)
            self.colatitude[:] = np.arctan2(np.sqrt(self.x**2 + self.y**2), self.z)

        # To perform the peak detection in 2D on a non-squared grid it is
        # necessary to know the neighboring points of each grid point.  The
        # Convex Hull of points on the sphere is equivalent to the Delauney
        # triangulation of the point set, which is what we are looking for.

        # Now we also want to compute the convex hull
        self.hull = sp.ConvexHull(self.cartesian.T)

        # and create an adjacency list
        adjacency = [set() for pt in range(self.n_points)]

        # Simplices contains all the triangles that form
        # the faces of the convex hull. We use this to find which
        # points are connected on the hull.
        for tri in self.hull.simplices:
            adjacency[tri[0]].add(tri[1])
            adjacency[tri[0]].add(tri[2])
            adjacency[tri[1]].add(tri[0])
            adjacency[tri[1]].add(tri[2])
            adjacency[tri[2]].add(tri[0])
            adjacency[tri[2]].add(tri[1])

        # convert to list of lists
        self.neighbors = [list(x) for x in adjacency]

    def apply(self, func, spherical=False):
        """
        Apply a function to every grid point
        """

        if spherical:
            self.values = func(self.azimuth, self.colatitude)
        else:
            self.values = func(self.x, self.y, self.z)

    def min_max_distance(self):
        """Compute some statistics on the distribution of the points"""

        min_dist = np.inf
        max_dist = 0

        dist = []

        for u in range(self.n_points):

            phi1, theta1 = self.spherical[:, u]

            for v in self.neighbors[u]:

                phi2, theta2 = self.spherical[:, v]

                d = great_circ_dist(1, theta1, phi1, theta2, phi2)

                dist.append(d)

                if d < min_dist:
                    min_dist = d

                if d > max_dist:
                    max_dist = d

        dist = np.array(dist)
        mean = dist.mean()
        median = np.median(dist)

        return min_dist, max_dist, mean, median

    def find_peaks(self, k=1):
        """
        Find the largest peaks on the grid
        """

        candidates = []

        # We start by looking at points whose neighbors all have lower values
        # than themselves
        for v in range(self.n_points):

            is_local_max = True

            for u in self.neighbors[v]:
                if self.values[u] > self.values[v]:
                    is_local_max = False
                    break

            if is_local_max:
                candidates.append(v)

        # Now sort and return k largest
        I = np.argsort(self.values[candidates])

        return [candidates[x] for x in I[-k:]]

    def regrid(self):
        """Regrid the non-uniform data on a regular mesh"""

        if self.values is None:
            warnings.warn("Cannont regrid: data missing.")
            return

        # First we need to interpolate the non-uniformly sampled data
        # onto a uniform grid
        from scipy.interpolate import griddata

        x = int(np.sqrt(self.n_points / 2))

        G = np.mgrid[-np.pi : np.pi : 2j * x, 0 : np.pi : 1j * x]
        spherical_grid = np.squeeze(G.reshape((2, 1, -1)))

        gridded = griddata(
            self.spherical.T,
            self.values,
            spherical_grid.T,
            method="nearest",
            fill_value=0.0,
        )
        dirty_img = gridded.reshape((2 * x, x))

        return G[0], G[1], gridded.reshape((2 * x, x))

    def plot(
        self,
        colatitude_ref=None,
        azimuth_ref=None,
        colatitude_recon=None,
        azimuth_recon=None,
        plotly=True,
        projection=True,
        points_only=False,
    ):

        if points_only:
            dirty_img = None
            dirty_grid_x = None
            dirty_grid_y = None
        else:
            dirty_grid_x, dirty_grid_y, dirty_img = self.regrid()

        ## Then, we just need to call Hanjie's routines
        from .plotters import sph_plot_diracs, sph_plot_diracs_plotly

        if projection:
            sph_plot_diracs(
                colatitude_ref,
                azimuth_ref,
                colatitude_recon,
                azimuth_recon,
                dirty_img=dirty_img,
                colatitude_grid=dirty_grid_y,
                azimuth_grid=dirty_grid_x,
            )

        if plotly:
            sph_plot_diracs_plotly(
                colatitude_ref,
                azimuth_ref,
                colatitude_recon,
                azimuth_recon,
                dirty_img=dirty_img,
                azimuth_grid=dirty_grid_x,
                colatitude_grid=dirty_grid_y,
            )

    def plot_old(self, plot_points=False, mark_peaks=0):
        """Plot the points on the sphere with their values"""

        from scipy import rand

        try:
            import matplotlib.colors as colors

            # from mpl_toolkits.mplot3d import Axes3D
            import mpl_toolkits.mplot3d as a3
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings

            warnings.warn("Matplotlib is required for plotting")
            return

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")

        if plot_points:
            ax.scatter(self.x, self.y, self.z, c="b", marker="o")

        if mark_peaks > 0:

            id = self.find_peaks(k=mark_peaks)
            s = 1.05
            ax.scatter(
                s * self.x[id], s * self.y[id], s * self.z[id], c="k", marker="o"
            )

        voronoi = sp.SphericalVoronoi(self.cartesian.T)
        voronoi.sort_vertices_of_regions()

        if self.values is not None:

            col_max = self.values.max()
            col_min = self.values.min()

            if col_min != col_max:
                col_map = (self.values - col_min) / (col_max - col_min)
            else:
                col_map = self.values / col_max

        else:

            col_map = np.zeros(self.n_points)

        cmap = plt.get_cmap("coolwarm")

        # plot the walls
        for v_ind, col in zip(voronoi.regions, col_map):
            triangle = a3.art3d.Poly3DCollection(
                [voronoi.vertices[v_ind]], alpha=1.0, linewidth=0.0
            )
            triangle.set_color(cmap(col))
            triangle.set_edgecolor("k")
            ax.add_collection3d(triangle)

        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
        ax.set_xlim([-1, 1])
        ax.set_ylim([-1, 1])
        ax.set_zlim([-1, 1])
